from langchain_core.documents import Document as LCDocument
import shortuuid

from rustic_ai.core.agents.commons.media import Document


class DocUtils:
    @staticmethod
    def from_langchain(langchain_document: LCDocument) -> Document:
        return Document(
            id=langchain_document.id or shortuuid.uuid(),
            metadata=langchain_document.metadata,
            content=langchain_document.page_content,
        )

    @staticmethod
    def to_langchain(document: Document) -> LCDocument:
        metadata = document.metadata or {}
        metadata["name"] = document.name
        metadata["mimetype"] = document.mimetype
        metadata["encoding"] = document.encoding

        return LCDocument(
            page_content=document.content,
            id=document.id,
            metadata=metadata,
        )



from typing import Callable, List, Literal, Optional, Union

from langchain_text_splitters import RecursiveCharacterTextSplitter
from pydantic import BaseModel

from rustic_ai.core.guild.agent_ext.depends import DependencyResolver
from rustic_ai.core.guild.agent_ext.depends.text_splitter import TextSplitter
from rustic_ai.core.utils import JsonDict


class RecursiveSplitterConf(BaseModel):
    """
    Configuration class for a recursive splitter.

    This class specifies the configuration for a recursive splitting algorithm
    used to divide text into smaller chunks based on specified rules. The class
    allows customization of chunk size, overlap, and other properties that define
    how the splitting process occurs.

    Attributes:
        chunk_size (int): The size of each text chunk.
        chunk_overlap (int): The overlap size between consecutive chunks.
        length_function (Callable[[str], int]): Function to calculate the length
            of a text. Default is the built-in `len` function.
        add_start_index (bool): Indicates whether to include the start index of
            each chunk in the output.
        strip_whitespace (bool): Specifies if leading and trailing whitespaces
            should be removed from chunks.
        separators (List[str]): A list of delimiters or separators to use for
            splitting text.
        keep_separator (Union[bool, Literal["start", "end"]]): Determines whether
            to retain the separator in the resulting chunks, and if so, whether
            to keep it at the start or end of the chunk.
        is_separator_regex (bool): Indicates if the separators provided should
            be treated as regular expressions.
        hf_tokenizer_model (Optional[str]): The name of a Hugging Face tokenizer
            model to be used for token-based splitting, if applicable.
            If specified, the unit for chunk_size is tokens instead of characters
            and tokenizer is used for the length_function
    """

    chunk_size: int = 1000
    chunk_overlap: int = 100
    length_function: Callable[[str], int] = len
    add_start_index: bool = False
    strip_whitespace: bool = True
    separators: List[str] = ["\n\n", "\n", " ", ""]
    keep_separator: Union[bool, Literal["start", "end"]] = True
    is_separator_regex: bool = False
    hf_tokenizer_model: Optional[str] = None


class RecursiveSplitter(TextSplitter):
    """
    Recursive text splitter that splits text into chunks of a specified size.
    """

    def __init__(self, conf: RecursiveSplitterConf):
        if conf.hf_tokenizer_model:
            from transformers import AutoTokenizer

            tokenizer = AutoTokenizer.from_pretrained(conf.hf_tokenizer_model)
            self.tokenizer = tokenizer
            self.splitter = RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
                tokenizer, **conf.model_dump(exclude={"hf_tokenizer_model", "length_function"})
            )
        else:
            self.splitter = RecursiveCharacterTextSplitter(
                separators=conf.separators,
                is_separator_regex=conf.is_separator_regex,
                keep_separator=conf.keep_separator,
                **conf.model_dump(exclude={"separators", "is_separator_regex", "keep_separator", "hf_tokenizer_model"}),
            )

    def split(self, text: str) -> list[str]:
        return self.splitter.split_text(text)


class RecursiveSplitterResolver(DependencyResolver[TextSplitter]):
    memoize_resolution: bool = False

    def __init__(self, conf: JsonDict = {}):
        super().__init__()
        splitter_conf = RecursiveSplitterConf.model_validate(conf)
        self.splitter = RecursiveSplitter(splitter_conf)

    def resolve(self, guild_id: str, agent_id: str) -> TextSplitter:
        return self.splitter



from typing import List, Tuple

from langchain_core.documents import Document as LCDocument
from langchain_core.indexing import UpsertResponse as LCUpsertResponse

from rustic_ai.core.guild.agent_ext.depends.vectorstore.vectorstore import (
    UpsertResponse,
    VectorSearchResult,
    VectorSearchResults,
)
from rustic_ai.langchain.utils.doc_utils import DocUtils


class VectorSearchUtils:
    @staticmethod
    def from_langchain_response(langchain_response: Tuple[LCDocument, float]) -> VectorSearchResult:  # pragma: no cover
        langchain_doc, score = langchain_response
        return VectorSearchResult(document=DocUtils.from_langchain(langchain_doc), score=score)

    @staticmethod
    def from_langchain_responses(
        query: str, langchain_responses: List[Tuple[LCDocument, float]]
    ) -> VectorSearchResults:
        results = [
            VectorSearchResult(document=DocUtils.from_langchain(langchain_doc), score=score)
            for langchain_doc, score in langchain_responses
        ]
        return VectorSearchResults(query=query, results=results)


class UpsertResponseUtils:
    @staticmethod
    def from_langchain(langchain_response: LCUpsertResponse) -> UpsertResponse:  # pragma: no cover
        return UpsertResponse(succeeded=langchain_response["succeeded"], failed=langchain_response["failed"])



from typing import Callable, Literal, Union

from langchain_text_splitters import CharacterTextSplitter
from pydantic import BaseModel

from rustic_ai.core.guild.agent_ext.depends import DependencyResolver
from rustic_ai.core.guild.agent_ext.depends.text_splitter import TextSplitter
from rustic_ai.core.utils import JsonDict


class CharacterSplitterConf(BaseModel):
    """
    Configuration for the CharacterSplitter.
    """

    chunk_size: int = 1000
    chunk_overlap: int = 100
    length_function: Callable[[str], int] = len
    keep_separator: Union[bool, Literal["start", "end"]] = False
    add_start_index: bool = False
    strip_whitespace: bool = True
    separator: str = "\n\n"
    is_separator_regex: bool = False


class CharacterSplitter(TextSplitter):
    """
    Character text splitter that splits text into chunks of a specified size.
    """

    def __init__(self, conf: CharacterSplitterConf):
        self.splitter = CharacterTextSplitter(
            separator=conf.separator,
            is_separator_regex=conf.is_separator_regex,
            **conf.model_dump(exclude={"separator", "is_separator_regex"}),
        )

    def split(self, text: str) -> list[str]:
        return self.splitter.split_text(text)


class CharacterSplitterResolver(DependencyResolver[TextSplitter]):
    memoize_resolution: bool = False

    def __init__(self, conf: JsonDict = {}):
        super().__init__()
        splitter_conf = CharacterSplitterConf.model_validate(conf)
        self.splitter = CharacterSplitter(splitter_conf)

    def resolve(self, guild_id: str, agent_id: str) -> TextSplitter:
        return self.splitter



from typing import List, Optional

from langchain_core.embeddings import Embeddings as LangchainEmbeddings
from langchain_openai import OpenAIEmbeddings as LangchainOAIE
from pydantic import BaseModel, Field

from rustic_ai.core.guild.agent_ext.depends import DependencyResolver
from rustic_ai.core.guild.agent_ext.depends.embeddings import Embeddings
from rustic_ai.core.utils import JsonDict

DEFAULT_MODEL: str = "text-embedding-ada-002"


class OpenAIEmbeddingConf(BaseModel):
    openai_api_version: Optional[str] = Field(default=None)
    """Automatically inferred from env var `OPENAI_API_VERSION` if not provided."""
    # to support Azure OpenAI Service custom endpoints
    openai_api_base: Optional[str] = Field(default=None)
    """Base URL path for API requests, leave blank if not using a proxy or service
        emulator."""
    # to support Azure OpenAI Service custom endpoints
    openai_api_type: Optional[str] = None
    # to support explicit proxy for OpenAI
    openai_proxy: Optional[str] = None
    embedding_ctx_length: int = 8191
    """The maximum number of tokens to embed at once."""
    openai_api_key: Optional[str] = Field(default=None)
    """Automatically inferred from env var `OPENAI_API_KEY` if not provided."""
    openai_organization: Optional[str] = Field(default=None)
    """Automatically inferred from env var `OPENAI_ORG_ID` if not provided."""


class OpenAIEmbeddings(Embeddings):
    """
    Embeddings backed by Open AI.
    """

    def __init__(self, model: str, deployment: str, conf: OpenAIEmbeddingConf):
        self.oaie = LangchainOAIE(
            model=model,
            deployment=deployment,
            **conf.model_dump(exclude_none=True, exclude_unset=True),
        )

    def embed(self, text: List[str]) -> List[List[float]]:
        return self.oaie.embed_documents(text)

    @property
    def langchain_embeddings(self) -> LangchainEmbeddings:
        return self.oaie


class OpenAIEmbeddingsResolver(DependencyResolver[Embeddings]):
    memoize_resolution: bool = False

    def __init__(
        self,
        model_name: str = DEFAULT_MODEL,
        deployment: Optional[str] = None,
        model_conf: JsonDict = {},
    ):
        super().__init__()
        mconf = OpenAIEmbeddingConf.model_validate(model_conf)

        self.embedding = OpenAIEmbeddings(
            model=model_name,
            deployment=deployment or model_name,
            conf=mconf,
        )

    def resolve(self, guild_id: str, agent_id: str) -> Embeddings:
        return self.embedding
